// some cut-and-paste and transcription from http://science.raphael.poss.name/rust-for-functional-programmers.html

// defining a bunch of functions, but only trying a few at a time.
#![allow(dead_code)]
// same with variables
#![allow(unused_variables)]

extern crate rand;

use rand::Rng;


// record types
struct Point {
    x: i32,
    y: i32
}

fn nullrecordtypetest() {
    let v = Point {x:1, y:2};
    let s = v.x + v.y;
}


// free type parameters

// according to the compiler, type parameters should be camelcase,
// not lowercase. tutorial page included 'a', 'b', and 't'. we'll change
// those to 'typeA', 'typeB', 'typeT'.
//
// wait, apparently it means upper camel case.
// 'TypeA', 'TypeB', 'TypeT'

type Pair<TypeA,TypeB> = (TypeA,TypeB);

// id<t> : |t| -> t
fn id<TypeT>(x: TypeT) -> TypeT { x }


// algebraic datatypes

// Option exists in the prelude, so MyOption
// ... and MyNone, MySome
enum MyOption<TypeT> {
    MyNone,
    MySome(TypeT)
}

// (x:MyOption<_>) does not work here, nor
// does (x:MyOption<TypeName>)
// using i32 to get it to compile
fn nullalgebraicdatatypes(x:MyOption<i32>) -> bool {
    // have to import the constructor names?
    use MyOption::MyNone;
    use MyOption::MySome;
    match x {
        MyNone     => false,
        MySome(_)  => true
    }
}


// lambda expressions & higher-order functions

// ||int,int| -> int, int| -> int
// fn ff(f:(|i32,i32|->i32), x:i32) -> i32 { f(x, x) }
// can't put bare function types in there, must make an
// ad-hoc named type and use that
fn ff<F: Fn(i32,i32) -> i32> (f:F, x:i32) -> i32 { f(x,x) }

// m2: |int| -> int
fn m2(n:i32) -> i32 { ff((|x,y| { x + y }), n) }

// trying some function type stuff

type F = Fn(i32,i32) -> i32;

// fn blergh : F(x,y) -- doesn't work


// type classes
trait Testable {
    fn test(&self) -> bool;
}

impl Testable for i32 {
    fn test(&self) -> bool {
        if *self == 0 { false }
        else { true }
    }
}

fn hellotest(x:i32) -> bool {
    x.test()
}


// operator overloading

enum Peano {
    PZero,
    Succ(Box<Peano>)
}

impl PartialEq for Box<Peano> {
    fn eq(&self, other:&Box<Peano>) -> bool {
        (self.deref() == other.deref())
    }
}

impl PartialEq for Peano {
    fn eq(&self, other:&Peano) -> bool {
        use Peano::PZero;
        use Peano::Succ;
        match (self, other) {
            (&PZero, &PZero)             => true,
            (&Succ(ref a), &Succ(ref b)) => (a == b),
            (_, _)                       => false
        }
    }
}

trait PartialEq {
    fn eq(&self, other:&Self) -> bool;
    fn ne(&self, other:&Self) -> bool
    { ! self.eq(other) }
}


// function types and definitions

// f : |int,int| -> int
fn f (x:i32, y:i32) -> i32 { x + y }

// fact : |int| -> int
// this will overflow the int
fn fact_rec (n:u64) -> u64 {
   if n == 1 { 1 }
   else { n * fact(n-1) }
}


// fact : |int| -> int
// the fold also overflows, but faster
fn fact (n:u64) -> u64 {
    (1..n+1).fold(1, |p,n| p*n)

    // (1..n+1) -- right-exclusive range interval
    // inclusive range is available as "experimental", and looks like: (1...n+1)
    // you can '#![feature(inclusive_range_syntax)]'
    // but 'feature' doesn't work on the stable releases, you need a nightly

    // fold - defined in the std::iter::Iterator, which is exposed
    // by the rust prelude: https://doc.rust-lang.org/std/prelude/

    // fold takes 2 arguments: initial value, and closure
    // |foo| expr -- this is a closure. can also be |foo| { block }
}


// effect-ful recursion

fn collatz(n:i32) {
    let v = match n % 2 {
        0 => n / 2,
        _ => 3 * n + 1
    };

    println!("{}", v);
    if v != 1 { collatz(v) }
}


// playground

fn main() {
    let start = rand::thread_rng().gen_range(1,101);

    println!("{}", start);
    collatz(start);

    // println!("fact: {}", fact(i64::from(start)));
    // there is no i -> u 'from'
    // println!("fact: {}", fact(start as u64));

    for i in 1..10 {
        println!("{}, {}, {}", i, fact_rec(i), m2(i as i32))
    }

    for i in 1..10 {
        println!("{}, {}", i, fact(i))
    }

}

